"""
React template
"""
import pathlib

import panel as pn
import param
from bokeh.themes import Theme as _BkTheme
from panel.template.theme import DarkTheme, DefaultTheme

from awesome_panel_extensions.frameworks.fast.templates import theme

from .grid_base_template import GridBaseTemplate


def _get_theme_from_query_args(default: str = "default") -> str:
    theme_arg = pn.state.session_args.get("theme", default)
    if isinstance(theme_arg, list):
        theme_arg = theme_arg[0].decode("utf-8")
        theme_arg = theme_arg.strip("'").strip('"')
    return theme_arg


class FastGridTemplate(GridBaseTemplate):
    """
    The FastTemplate is build on top of Fast.design and the React Grid Layout.
    """

    enable_theme_toggle = param.Boolean(
        default=True, doc="If True a switch is to toggle the Theme. Default is True"
    )

    def __init__(self, **params):
        if "theme" not in params:
            _theme = "dark"
            if params.get("enable_theme_toggle", self.param.enable_theme_toggle.default):
                _theme = _get_theme_from_query_args(default=_theme)
            if _theme == "default":
                params["theme"] = FastDefaultTheme
            else:
                params["theme"] = FastDarkTheme

        super().__init__(**params)

        if "header_color" not in params:
            self.header_color = self.theme.style.header_color
        if "header_background" not in params:
            self.header_background = self.theme.style.header_background
        self._update_special_render_vars()

    def _update_special_render_vars(self):
        self._render_variables["css_base"] = pathlib.Path(self._css).read_text()
        self._render_variables["css_theme"] = pathlib.Path(self.theme.css).read_text()
        self._render_variables["js"] = pathlib.Path(self._js).read_text()
        if self.theme is FastDarkTheme:
            self._render_variables["theme"] = "dark"
        else:
            self._render_variables["theme"] = "default"
        self._render_variables["style"] = self.theme.style
        self._render_variables["enable_theme_toggle"] = self.enable_theme_toggle

    _css = pathlib.Path(__file__).parent / "fast_grid_template.css"
    _js = pathlib.Path(__file__).parent / "fast_grid_template.js"

    _template = pathlib.Path(__file__).parent / "fast_grid_template.html"


class FastDefaultTheme(DefaultTheme):
    """The Default Theme of the Fast templates"""

    css = param.Filename(default=pathlib.Path(__file__).parent / "default.css")

    _template = FastGridTemplate

    style = param.ClassSelector(class_=theme.FastStyle, default=theme.DEFAULT_STYLE)

    bokeh_theme = param.ClassSelector(
        class_=(_BkTheme, str), default=_BkTheme(json=theme.DEFAULT_THEME)
    )


class FastDarkTheme(DarkTheme):
    """The Dark Theme of the Fast Templates"""

    css = param.Filename(default=pathlib.Path(__file__).parent / "dark.css")

    _template = FastGridTemplate

    style = param.ClassSelector(class_=theme.FastStyle, default=theme.DARK_STYLE)

    bokeh_theme = param.ClassSelector(
        class_=(_BkTheme, str), default=_BkTheme(json=theme.DARK_THEME)
    )



"""In this module we define the BaseModel.

The BaseModel adds ordering by the name parameter to a Class"""
import param


class BaseModel(param.Parameterized):
    """The BaseModel adds ordering by the name parameter to a Class"""

    def __lt__(self, other):
        if hasattr(other, "name"):
            return self.name.casefold() < other.name.casefold()
        return True

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name
        return False

    def __str__(
        self,
    ):
        return self.name

    def __repr__(
        self,
    ):
        return self.name



"""The GridBaseTemplate is a copy of the ReactTemplate"""
import param
from panel.depends import depends
from panel.layout import Card, GridSpec
from panel.template.base import BasicTemplate

# pylint: disable=unused-variable, invalid-name, line-too-long


class GridBaseTemplate(BasicTemplate):
    """The GridBaseTemplate is a copy of the ReactTemplate"""

    compact = param.ObjectSelector(default=None, objects=[None, "vertical", "horizontal", "both"])

    cols = param.Dict(default={"lg": 12, "md": 10, "sm": 6, "xs": 4, "xxs": 2})

    breakpoints = param.Dict(default={"lg": 1200, "md": 996, "sm": 768, "xs": 480, "xxs": 0})

    main = param.ClassSelector(
        class_=GridSpec,
        constant=True,
        doc="""
        A list-like container which populates the main area.""",
    )

    row_height = param.Integer(default=50)

    enable_theme_selection = param.Boolean(
        default=True,
        doc="""If True it will be possible for the
    user to toggle between the 'default' and 'dark' theme. Default""",
    )

    _modifiers = {Card: {"children": {"margin": (20, 20)}, "margin": (10, 5)}}

    _resources = {
        "js": {
            "react": "https://unpkg.com/react@16/umd/react.development.js",
            "react-dom": "https://unpkg.com/react-dom@16/umd/react-dom.development.js",
            "babel": "https://unpkg.com/babel-standalone@latest/babel.min.js",
            "react-grid": "https://cdnjs.cloudflare.com/ajax/libs/react-grid-layout/1.1.1/react-grid-layout.min.js",
        },
        "css": {},
    }

    def __init__(self, **params):
        if "main" not in params:
            params["main"] = GridSpec(ncols=12, mode="override")
        super().__init__(**params)
        self._update_render_vars()

    def _update_render_items(self, event):
        super()._update_render_items(event)
        if event.obj is not self.main:
            return
        layouts = []
        for i, ((y0, x0, y1, x1), v) in enumerate(self.main.objects.items()):
            if x0 is None:
                x0 = 0
            if x1 is None:
                x1 = 12
            if y0 is None:
                y0 = 0
            if y1 is None:
                y1 = self.main.nrows
            layouts.append({"x": x0, "y": y0, "w": x1 - x0, "h": y1 - y0, "i": str(i + 1)})
        self._render_variables["layouts"] = {"lg": layouts, "md": layouts}

    @depends("cols", "breakpoints", "row_height", "compact", watch=True)
    def _update_render_vars(self):
        self._render_variables["breakpoints"] = self.breakpoints
        self._render_variables["cols"] = self.cols
        self._render_variables["rowHeight"] = self.row_height
        self._render_variables["compact"] = self.compact



"""The Material Extension should be included in your app layout of the value of `.object`
included in your Template"""
# pylint: enable=line-too-long
import panel as pn
import param

# pylint: disable=line-too-long
# <script src="https://cdn.jsdelivr.net/gh/marcskovmadsen/awesome-panel@be59521090b7c9d9ba5eb16e936034e412e2c86b/assets/js/mwc.bundled.js"></script>

_MWC_SCRIPTS = """
    <script src="https://unpkg.com/@material/mwc-button@0.18.0?module" type="module"></script>
    <script src="https://unpkg.com/@material/mwc-select@0.18.0?module" type="module"></script>
    <script src="https://unpkg.com/@material/mwc-list@0.18.0?module" type="module"></script>
    <script src="https://unpkg.com/@material/mwc-slider@0.18.0?module" type="module"></script>
    <script src="https://unpkg.com/@material/mwc-linear-progress@0.18.0?module" type="module"></script>
    <script src="https://unpkg.com/@material/mwc-circular-progress@0.18.0?module" type="module"></script>
"""
_MWC_FONTS = (
    '<link href="https://fonts.googleapis.com/css?family=Roboto:300,400,500" rel="stylesheet">'
    '<link href="https://fonts.googleapis.com/css?family=Material+Icons&display=block" rel="stylesheet">'
)
_MDC_STYLE_SHEET = '<link rel="stylesheet" href="https://unpkg.com/material-components-web@latest/dist/material-components-web.min.css">'
_MWC_ALL = _MWC_FONTS + _MDC_STYLE_SHEET + _MWC_SCRIPTS


class Extension(pn.pane.HTML):  # pylint: disable=too-few-public-methods
    """The Material Extension should be included in your app layout of the value of `.object`
    included in your Template"""

    object = param.Parameter(
        default=None,
        doc="""
        The .js and .css dependencies to include.""",
        constant=True,
    )
    # In order to not be selected by the `pn.panel` selection process
    # Cf. https://github.com/holoviz/panel/issues/1494#issuecomment-663219654
    priority = 0
    # The _rename dict is used to keep track of Panel parameters to sync to Bokeh properties.
    # As value is not a property on the Bokeh model we should set it to None
    _rename = {
        **pn.pane.HTML._rename,
        "value": None,
    }

    def __init__(self, **params):
        params["object"] = _MWC_ALL
        params["height"] = 0
        params["width"] = 0
        params["sizing_mode"] = "fixed"
        params["margin"] = 0
        super().__init__(**params)



"""The TemplateSettings contains a collecation of user definable settings"""
import panel as pn
import param

from awesome_panel_extensions._shared.logger import get_logger
from awesome_panel_extensions.models.icon import Icon
from awesome_panel_extensions.widgets.button import AwesomeButton

logger = get_logger(__name__)

PARAMETERS = ["theme", "template"]

# pylint: disable=line-too-long
DEFAULT_ICON = Icon(
    name="Sun",
    value="""<svg width="1em" height="1em" viewBox="0 0 16 16" class="bi bi-sun" fill="currentColor" xmlns="http://www.w3.org/2000/svg"><path d="M3.5 8a4.5 4.5 0 1 1 9 0 4.5 4.5 0 0 1-9 0z"/><path fill-rule="evenodd" d="M8.202.28a.25.25 0 0 0-.404 0l-.91 1.255a.25.25 0 0 1-.334.067L5.232.79a.25.25 0 0 0-.374.155l-.36 1.508a.25.25 0 0 1-.282.19l-1.532-.245a.25.25 0 0 0-.286.286l.244 1.532a.25.25 0 0 1-.189.282l-1.509.36a.25.25 0 0 0-.154.374l.812 1.322a.25.25 0 0 1-.067.333l-1.256.91a.25.25 0 0 0 0 .405l1.256.91a.25.25 0 0 1 .067.334L.79 10.768a.25.25 0 0 0 .154.374l1.51.36a.25.25 0 0 1 .188.282l-.244 1.532a.25.25 0 0 0 .286.286l1.532-.244a.25.25 0 0 1 .282.189l.36 1.508a.25.25 0 0 0 .374.155l1.322-.812a.25.25 0 0 1 .333.067l.91 1.256a.25.25 0 0 0 .405 0l.91-1.256a.25.25 0 0 1 .334-.067l1.322.812a.25.25 0 0 0 .374-.155l.36-1.508a.25.25 0 0 1 .282-.19l1.532.245a.25.25 0 0 0 .286-.286l-.244-1.532a.25.25 0 0 1 .189-.282l1.508-.36a.25.25 0 0 0 .155-.374l-.812-1.322a.25.25 0 0 1 .067-.333l1.256-.91a.25.25 0 0 0 0-.405l-1.256-.91a.25.25 0 0 1-.067-.334l.812-1.322a.25.25 0 0 0-.155-.374l-1.508-.36a.25.25 0 0 1-.19-.282l.245-1.532a.25.25 0 0 0-.286-.286l-1.532.244a.25.25 0 0 1-.282-.189l-.36-1.508a.25.25 0 0 0-.374-.155l-1.322.812a.25.25 0 0 1-.333-.067L8.203.28zM8 2.5a5.5 5.5 0 1 0 0 11 5.5 5.5 0 0 0 0-11z"/></svg>""",
    fill_color="#FF9800",
)

DARK_ICON = Icon(
    name="Github",
    value="""<svg width="1em" height="1em" viewBox="0 0 16 16" class="bi bi-moon" fill="currentColor" xmlns="http://www.w3.org/2000/svg"><path fill-rule="evenodd" d="M14.53 10.53a7 7 0 0 1-9.058-9.058A7.003 7.003 0 0 0 8 15a7.002 7.002 0 0 0 6.53-4.47z"/></svg>""",
    fill_color="black",
)
# pylint: enable=line-too-long
class TemplateSettings(param.Parameterized):
    """Collection of user definable settings"""

    toggle_theme = param.Action(label="")
    theme = param.ObjectSelector("default", objects=["default", "dark"])
    template = param.ObjectSelector(
        "material", objects=["bootstrap", "golden", "material", "react", "vanilla"]
    )

    settings_panel = param.Parameter()
    js_panel = param.Parameter()
    view = param.Parameter()

    def __init__(self, parameters=None, **params):
        logger.info("Initializing Settings %s", pn.state.session_args)
        self._update_param_from_session_args("theme", params)
        self._update_param_from_session_args("template", params)
        super().__init__(**params)

        if not parameters:
            parameters = PARAMETERS
        parameters = [*parameters]

        if "theme" in parameters:
            parameters.append("toggle_theme")
        parameters = [p for p in parameters if p != "theme"]

        self.toggle_theme = self._toggle_theme

        self._create_panels_and_view(parameters)

    def _create_panels_and_view(self, parameters):
        if self.theme == "dark":
            icon = DARK_ICON
        else:
            icon = DEFAULT_ICON

        self.settings_panel = pn.Param(
            self,
            parameters=parameters,
            widgets={
                "template": {"width": 100, "sizing_mode": "fixed"},
                "toggle_theme": {
                    "type": AwesomeButton,
                    "width": 50,
                    "sizing_mode": "fixed",
                    "icon": icon,
                },
            },
            default_layout=pn.Row,
            show_labels=False,
            show_name=False,
            sizing_mode="fixed",
            width=200,
            css_classes=["template-settings"],
        )

        if self.js_panel:
            self.view = self.settings_panel
        else:
            self.js_panel = pn.pane.HTML(height=0, width=0, margin=0, sizing_mode="fixed")
            self.view = pn.Row(self.settings_panel, self.js_panel, sizing_mode="fixed", width=200)

    def _update_param_from_session_args(self, parameter, params):
        if parameter in pn.state.session_args:
            value = pn.state.session_args[parameter][0].decode("utf-8")
            value = value.strip("'").strip('"').replace("%22", "")
            if value in self.param[parameter].objects:
                params[parameter] = value

    def _toggle_theme(self, *_):
        if self.theme == "default":
            self.theme = "dark"
        else:
            self.theme = "default"

    @param.depends("theme", "template", watch=True)
    def _change_query_params(self, *_):
        pn.state.location.update_query(template=self.template, theme=self.theme)
        logger.info("reload to %s", pn.state.location.query_params)
        self.js_panel.object = "<script>location.reload()</script>"


if __name__.startswith("bokeh"):
    TemplateSettings().view.servable()



"""The DEFAULT_STYLE and DARK_STYLE contains the different colors and icons used the style the Fast
Templates"""

from typing import Dict

import param

COLLAPSED_SVG_ICON = """
<svg style="stroke: #E62F63" width="18" height="18" viewBox="0 0 18 18" fill="none" xmlns="http://www.w3.org/2000/svg" slot="collapsed-icon">
            <path d="M15.2222 1H2.77778C1.79594 1 1 1.79594 1 2.77778V15.2222C1 16.2041 1.79594 17 2.77778 17H15.2222C16.2041 17 17 16.2041 17 15.2222V2.77778C17 1.79594 16.2041 1 15.2222 1Z" stroke-linecap="round" stroke-linejoin="round"></path>
            <path d="M9 5.44446V12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
            <path d="M5.44446 9H12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
        </svg>
""".replace(
    "\n", ""
)

EXPANDED_SVG_ICON = """
<svg style="stroke: #E62F63" width="18" height="18" viewBox="0 0 18 18" fill="none" xmlns="http://www.w3.org/2000/svg" slot="expanded-icon">
    <path d="M15.2222 1H2.77778C1.79594 1 1 1.79594 1 2.77778V15.2222C1 16.2041 1.79594 17 2.77778 17H15.2222C16.2041 17 17 16.2041 17 15.2222V2.77778C17 1.79594 16.2041 1 15.2222 1Z" stroke-linecap="round" stroke-linejoin="round"></path>
    <path d="M5.44446 9H12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
</svg>
""".replace(
    "\n", ""
)

FAST_DARK_100 = "rgb(48,48,48)"
FAST_DARK_75 = "rgb(57,57,57)"
FAST_DARK_50 = "rgb(66,66,66)"
FAST_DARK_25 = "rgb(77,77,77)"
# self.neutral_foreground_rest = "rgb(236,236,236)"


class FastStyle(param.Parameterized):
    """The FastStyle class provides the different colors and icons used to style the Fast
    Templates"""

    background_color = param.Color(default="#ffffff")
    color = param.Color(default="#00aa41")
    accent_base_color = param.Color("#1B5E20")
    neutral_fill_card_rest = param.Color("#F7F7F7")
    neutral_focus = param.Color("#888888")
    neutral_foreground_rest = param.Color("#2B2B2B")
    expanded_icon = param.String(EXPANDED_SVG_ICON)
    collapsed_icon = param.String(COLLAPSED_SVG_ICON)
    font = param.String(default="Open Sans, sans-serif")
    header_color = param.Color("#ffffff")
    header_background = param.Color("#1B5E20")

    def create_bokeh_theme(self) -> Dict:
        """Returns a custom bokeh theme based on the style parameters

        Returns:
            Dict: A Bokeh Theme
        """

        return {
            "attrs": {
                "Figure": {
                    "background_fill_color": self.background_color,
                    "border_fill_color": self.neutral_fill_card_rest,
                    "outline_line_color": self.neutral_focus,
                    "outline_line_alpha": 0.5,
                    "outline_line_width": 1,
                },
                "Grid": {"grid_line_color": self.neutral_focus, "grid_line_alpha": 0.25},
                "Axis": {
                    "major_tick_line_alpha": 0,
                    "major_tick_line_color": self.neutral_foreground_rest,
                    "minor_tick_line_alpha": 0,
                    "minor_tick_line_color": self.neutral_foreground_rest,
                    "axis_line_alpha": 0,
                    "axis_line_color": self.neutral_foreground_rest,
                    "major_label_text_color": self.neutral_foreground_rest,
                    "major_label_text_font": self.font,
                    "major_label_text_font_size": "1.025em",
                    "axis_label_standoff": 10,
                    "axis_label_text_color": self.neutral_foreground_rest,
                    "axis_label_text_font": self.font,
                    "axis_label_text_font_size": "1.25em",
                    "axis_label_text_font_style": "normal",
                },
                "Legend": {
                    "spacing": 8,
                    "glyph_width": 15,
                    "label_standoff": 8,
                    "label_text_color": self.neutral_foreground_rest,
                    "label_text_font": self.font,
                    "label_text_font_size": "1.025em",
                    "border_line_alpha": 0,
                    "background_fill_alpha": 0.25,
                    "background_fill_color": FAST_DARK_75,
                },
                "ColorBar": {
                    "title_text_color": self.neutral_foreground_rest,
                    "title_text_font": self.font,
                    "title_text_font_size": "1.025em",
                    "title_text_font_style": "normal",
                    "major_label_text_color": self.neutral_foreground_rest,
                    "major_label_text_font": self.font,
                    "major_label_text_font_size": "1.025em",
                    "background_fill_color": FAST_DARK_75,
                    "major_tick_line_alpha": 0,
                    "bar_line_alpha": 0,
                },
                "Title": {
                    "text_color": self.neutral_foreground_rest,
                    "text_font": self.font,
                    "text_font_size": "1.15em",
                },
            }
        }


DEFAULT_STYLE = FastStyle()
DARK_STYLE = FastStyle(
    background_color="#181818",
    color="#ffffff",
    neutral_fill_card_rest="#212121",
    neutral_focus="#717171",
    neutral_foreground_rest="#e5e5e5",
)
DEFAULT_THEME = DEFAULT_STYLE.create_bokeh_theme()
DARK_THEME = DARK_STYLE.create_bokeh_theme()



"""Implementation of the PerspectiveViewer Web Component"""
from enum import Enum
from typing import List

import panel as pn
import param

from awesome_panel_extensions.bokeh_extensions.perspective_viewer import (
    PerspectiveViewer as _BkPerspectiveViewer,
)
from awesome_panel_extensions.widgets.dataframe_base import DataFrameWithStreamAndPatchBaseWidget

# This is need to be able to use Perspective in notebook via pn.extension("perspective")
pn.extension._imports[  # pylint: disable=protected-access
    "perspective"
] = "awesome_panel_extensions.bokeh_extensions.perspective_viewer"

DEFAULT_THEME = "material"
THEMES_MAP = {
    "material": "perspective-viewer-material",
    "material-dark": "perspective-viewer-material-dark",
    "material-dense": "perspective-viewer-material-dense",
    "material-dense-dark": "perspective-viewer-material-dense-dark",
    "vaporwave": "perspective-viewer-vaporwave",
}
THEMES = [*THEMES_MAP.keys()]
# Hack: When the user drags some of the columns, then the class attribute contains "dragging" also.
CSS_CLASS_MAP = {v: k for k, v in THEMES_MAP.items()}
DEFAULT_CSS_CLASS = THEMES_MAP[DEFAULT_THEME]

# Source: https://github.com/finos/perspective/blob/e23988b4b933da6b90fd5767d059a33e70a2493e/python/perspective/perspective/core/plugin.py#L49 # pylint: disable=line-too-long
class Plugin(Enum):
    """The plugins (grids/charts) available in Perspective.  Pass these into
    the `plugin` arg in `PerspectiveWidget` or `PerspectiveViewer`.
    """

    HYPERGRID = "hypergrid"  # hypergrid
    GRID = "datagrid"  # hypergrid
    YBAR_D3 = "d3_y_bar"  # d3fc
    XBAR_D3 = "d3_x_bar"  # d3fc
    YLINE_D3 = "d3_y_line"  # d3fc
    YAREA_D3 = "d3_y_area"  # d3fc
    YSCATTER_D3 = "d3_y_scatter"  # d3fc
    XYSCATTER_D3 = "d3_xy_scatter"  # d3fc
    TREEMAP_D3 = "d3_treemap"  # d3fc
    SUNBURST_D3 = "d3_sunburst"  # d3fc
    HEATMAP_D3 = "d3_heatmap"  # d3fc
    CANDLESTICK = "d3_candlestick"  # d3fc
    CANDLESTICK_D3 = "d3_candlestick"  # d3fc
    OHLC = "d3_ohlc"  # d3fc
    OHLC_D3 = "d3_ohlc"  # d3fc

    @staticmethod
    def options() -> List:
        """Returns the list of options of the PerspectiveViewer, like Hypergrid, Grid etc.

        Returns:
            List: [description]
        """
        return list(c.value for c in Plugin)


class PerspectiveViewer(DataFrameWithStreamAndPatchBaseWidget):  # pylint: disable=abstract-method
    """The PerspectiveViewer widget enables exploring large tables of data"""

    _widget_type = _BkPerspectiveViewer
    plugin = param.ObjectSelector(
        Plugin.GRID.value,
        objects=Plugin.options(),
        doc="The name of a plugin to display the data. For example hypergrid or d3_xy_scatter.",
    )
    theme = param.ObjectSelector(
        DEFAULT_THEME,
        objects=THEMES,
        doc="The style of the PerspectiveViewer. For example material-dark",
    )
    columns = param.List(
        None, doc='A list of source columns to show as columns. For example ["x", "y"]'
    )
    # We don't expose this as it is not documented
    # parsed_computed_columns = param.List(
    #     None,
    #     doc='A list of parsed computed columns. For example [{"name":"x+y","func":"add","inputs":["x","y"]}]',
    # )
    computed_columns = param.List(
        None,
        doc='A list of computed columns. For example [""x"+"index""]',
    )
    column_pivots = param.List(
        None, doc='A list of source columns to pivot by. For example ["x", "y"]'
    )
    row_pivots = param.List(
        None, doc='A list of source columns to group by. For example ["x", "y"]'
    )
    aggregates = param.Dict(None, doc='How to aggregate. For example {x: "distinct count"}')
    sort = param.List(None, doc='How to sort. For example[["x","desc"]]')
    filters = param.List(
        None, doc='How to filter. For example [["x", "<", 3],["y", "contains", "abc"]]'
    )

    # I set this to something > 0. Otherwise the PerspectiveViewer widget will have a height of 0px
    # It will appear as if it does not work.
    height = param.Integer(default=300, bounds=(0, None))

    def __init__(self, **params):
        super().__init__(**params)

        self._set_source()



"""The FastSwitchGroup enables using the `fast-switch` with Bokeh."""
from bokeh.core.properties import Bool, String
from bokeh.models.widgets import CheckboxGroup as _BkCheckboxGroup


class FastSwitchGroup(_BkCheckboxGroup):
    """The FastSwitchGroup enables using the `fast-switch` with Bokeh."""

    readonly = Bool()
    checked_message = String()
    unchecked_message = String()
